# -*- coding:utf-8 -*-
"""
Created on the 05/01/18
@author: Nicolas Thiebaut
@email: nkthiebaut@gmail.com
"""
import os
from zeugma.logger import PACKAGE_LOGGER as logger

MODELS_DIR = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), 'zeugma_embeddings')
if not os.path.exists(MODELS_DIR):
    logger.info(MODELS_DIR + ' not found, creating it.')
    os.mkdir(MODELS_DIR)

OOV_TAG = '<oov>'  # Out-Of-Vocabulary tag for rare words



# -*- coding:utf-8 -*-
"""
Created on the 05/01/18
@author: Nicolas Thiebaut
@email: nkthiebaut@gmail.com
"""

from .embeddings import Word2VecTransformer, GloVeTransformer
from .texttransformers import RareWordsTagger, ItemSelector, TextStats



# -*- coding:utf-8 -*-
"""
Created on the 05/01/18
@author: Nicolas Thiebaut
@email: nkthiebaut@gmail.com
"""

from abc import ABCMeta, abstractmethod
from functools import reduce
import gzip
from multiprocessing import cpu_count
import os
import shutil
import urllib
import zipfile

from gensim.models import KeyedVectors, Word2Vec
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from zeugma.logger import PACKAGE_LOGGER as logger
from zeugma.conf import MODELS_DIR


class EmbeddingTransformer(BaseEstimator, TransformerMixin, metaclass=ABCMeta):
    """ Abstract text vectorizer class """

    def __init__(self, model_path=None, trainable=False, **kwargs):
        self.trainable = trainable
        self.model_path = model_path
        self.params = dict(kwargs)
        if trainable:
            for method in ['train', 'save']:
                if not hasattr(self, method):
                    raise NotImplementedError('trainable is set to True but ' +\
                            self.__class__.__name__ + ' does not implement ' + \
                                         method + ' method.')
        else:
            if model_path is None:
                logger.info('Embedding is not trainable and model_path not' +\
                            'specified, using default model location:' + MODELS_DIR)
                model_path = MODELS_DIR
            if not hasattr(self, 'load_pretrained_model'):
                raise NotImplementedError(self.__class__.__name__ + \
                                          ' does not support pretrained models.')
            else:
                if not os.path.exists(model_path):
                    raise FileNotFoundError(self.__class__.__name__ +
                                            ' model file not found')
            self.model = self.load_pretrained_model()

    @abstractmethod
    def transform_sentence(self, text):
        """ Child classes have to implement this method that will be used
        to transform corpora """
        pass

    def fit(self, x, y=None):
        """ Has to define fit method to conform scikit-learn Transformer
        definition and integrate a sklearn.Pipeline object """
        if self.trainable:
            self.train(x)
            if self.model_path is not None:
                self.save()
        return self

    def transform(self, texts):
        """ Transform corpus from single text transformation method """
        if hasattr(self, 'transform_sentence'):
            return np.array([self.transform_sentence(t) for t in texts])
        else:
            raise NotImplementedError()

    @abstractmethod
    def load_pretrained_model(self):
        """ Child classes have to implement this method that will be used
        to load a saved model """
        pass

    def train(self, corpus):
        """ Embedding training method for trainable EmbeddingTransformers """
        raise NotImplementedError('Training is not yet supported for ' +\
                                  self.__class__.__name__)

    def save(self):
        """ Method to save EmbeddingTransformers when trainable """
        raise NotImplementedError('Saving trained embeddings is not yet ' +\
                                  'supported for ' + self.__class__.__name__)


W2V_EMBEDDINGS_URL = "https://github.com/eyaler/word2vec-slim/raw/master/" +\
                     "GoogleNews-vectors-negative300-SLIM.bin.gz"


class Word2VecTransformer(EmbeddingTransformer):
    """ Word2Vec embeddings class, transforms a corpus to its w2v
    representation matrix

    Notes
    -----
    Google news pre-trained Word2vec download link:
    https://drive.google.com/file/d/0B7XkCwpI5KDYNlNUTTlSS21pQmM/edit?usp=sharing
    Or the lighter version (300k words instead of 3M) here:
    https://github.com/eyaler/word2vec-slim/raw/master/GoogleNews-vectors-negative300-SLIM.bin.gz
    """
    def load_pretrained_model(self):
        """ Load a pre-trained word2vec model """
        if self.model_path.endswith('.bin'):
            w2v = KeyedVectors.load_word2vec_format(self.model_path, binary=True)
        elif self.model_path.endswith('.vec'):
            w2v = KeyedVectors.load_word2vec_format(self.model_path, binary=False)
        else:
            raise NameError('Unknown file extension.')
        return w2v

    def train(self, corpus):
        default_params = dict(size=100, window=2, min_count=5,
                              workers=cpu_count())
        params = {k: self.params.get(k, default) for k, default in
                  default_params.items()}
        x = np.array([text.split() for text in corpus])
        self.model = Word2Vec(x, **params)

    def save(self):
        self.model.save(self.model_path)

    def transform_sentence(self, text):
        """ Compute mean w2v vector for the input text"""
        def preprocess_text(self, text):
            """ Prepare text for Gensim model, excluding unknown words"""
            if not isinstance(text, list):
                if not isinstance(text, str):
                    raise TypeError
                text = text.split()
            return list(filter(lambda x: x in self.model.wv.vocab, text))
        text = preprocess_text(self, text)
        if not text:
            return np.zeros(self.model.vector_size)
        return np.mean(self.model.wv[text], axis=0)
    
    @staticmethod
    def download_embeddings(model_path=MODELS_DIR + os.sep, url=W2V_EMBEDDINGS_URL,
                            outfile=None):
        """ Download Word2vec pre-computed embeddings from Eyaler github repo """
        gz_file = os.path.join(os.path.dirname(model_path),
                               'GoogleNews-vectors-negative300.bin')
        urllib.request.urlretrieve(url, gz_file)
        if outfile is None:
            outfile = gz_file[:-3]
        with gzip.open(gz_file, 'rb') as f_in, open(outfile, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)


GLOVE_EMBEDDINGS_URL = "http://nlp.stanford.edu/data/glove.6B.zip"


class GloVeTransformer(EmbeddingTransformer):
    """ Glove embeddings class, transforms a corpus to its Glove
    representation matrix"""

    def load_pretrained_model(self):
        """ Load a pre-trained GloVe model """
        if self.model_path.endswith('.txt'):
            embeddings_dict = dict()
            with open(self.model_path, encoding="utf8") as glove_stream:
                for line in glove_stream:
                    values = line.split()
                    word = values[0]
                    value = np.asarray(values[1:], dtype='float32')
                    embeddings_dict[word] = value
        else:
            raise NameError('Unknown file extension.')
        return embeddings_dict

    def transform_sentence(self, text):
        """ Return the mean of the words embeddings """
        size = len(self.model['the'])
        embeddings = (self.model.get(w, np.zeros(size)) for w in text)
        text_vector = reduce(np.add, embeddings, np.zeros(size))
        return text_vector / len(text)

    @staticmethod
    def download_glove_embeddings(model_path=MODELS_DIR + os.sep,
                                  url=GLOVE_EMBEDDINGS_URL):
        """ Download GloVe pre-computed embeddings from Stanford website """
        model_dir = os.path.dirname(model_path)
        zip_file = os.path.join(model_dir, 'glove.6B.zip')
        urllib.request.urlretrieve(url, zip_file)
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(model_dir)

    #TODO: add GloVe training with the glove_python library



# -*- coding:utf-8 -*-
"""
Created on the 01/26/2018
@author: Nicolas Thiebaut
@email: nicolas@visage.jobs
"""
from collections import defaultdict
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import FunctionTransformer
from zeugma.conf import OOV_TAG
from zeugma.logger import PACKAGE_LOGGER as logger


class RareWordsTagger(BaseEstimator, TransformerMixin):
    """ Replace rare words with a token in a corpus (list of strings) """

    def __init__(self, min_count, oov_tag=OOV_TAG):
        self.min_count = min_count
        self.oov_tag = oov_tag
        self.frequencies = defaultdict(int)
        self.kept_words = None

    def fit(self, texts, y=None):
        all_tokens = (token for t in texts for token in t.split())
        for w in all_tokens:
            self.frequencies[w] += 1
        logger.info('Vocabulary size before rare words tagging: ' +
                    str(len(self.frequencies)))
        self.kept_words = {word for word, frequency in self.frequencies.items()
                           if frequency >= self.min_count}
        logger.info('Vocabulary size after rare words tagging: ' +
                    str(len(self.kept_words)))
        return self

    def transform(self, texts):
        texts = [' '.join((w if w in self.kept_words else self.oov_tag
                           for w in t.split()))
                 for t in texts]
        return texts


class ItemSelector(BaseEstimator, TransformerMixin):
    """For data grouped by feature, select subset of data at a provided key.

    The data is expected to be stored in a 2D data structure, where the first
    index is over features and the second is over samples.

    Parameters
    ----------
    key : hashable, required
        The key corresponding to the desired value in a mappable.
    """
    def __init__(self, key):
        self.key = key

    def fit(self, x, y=None):
        """ Necessary fit method to include transformer in a sklearn.Pipeline"""
        return self

    def transform(self, data_dict):
        """ Return selected items """
        return data_dict[self.key]


class TextStats(FunctionTransformer):
    """Extract features from each document for DictVectorizer"""
    def __init__(self):
        def extract_stats(corpus):
            return [{'length': len(text),
                    'num_sentences': text.count('.')}
                    for text in corpus]
        super().__init__(extract_stats, validate=False)



# -*- coding:utf-8 -*-
"""
Created on the 01/25/2018
@author: Nicolas Thiebaut
@email: nicolas@visage.jobs
"""
import logging

from logging import StreamHandler
from logging import Formatter

LOG_FORMAT = (
    "%(asctime)s [%(levelname)s]: %(message)s in %(pathname)s:%(lineno)d")

DEFAULT_LOG_LEVEL = logging.DEBUG


DEFAULT_HANDLER = StreamHandler()
DEFAULT_HANDLER.setFormatter(Formatter(LOG_FORMAT))

PACKAGE_LOGGER = logging.getLogger("zeugma")
PACKAGE_LOGGER.setLevel(DEFAULT_LOG_LEVEL)
PACKAGE_LOGGER.addHandler(DEFAULT_HANDLER)



#!/usr/bin/env python

from distutils.core import setup

setup(name='zeugma',
      packages=['zeugma'],
      version='0.27',
      license='MIT',
      description='Unified framework for word embeddings (Word2Vec, GloVe, FastText, ...) compatible with scikit-learn Pipeline',
      author='Nicolas Thiebaut',
      author_email='nkthiebaut@gmail.com',
      url='https://github.com/nkthiebaut',
      download_url='https://github.com/nkthiebaut/zeugma/archive/0.27.tar.gz',
      keywords=['embeddings'],
      classifiers=[],
      setup_requires=[
          'numpy>=1.13.3',
          'Cython>=0.27.3',
      ],
      install_requires=[
          'numpy>=1.13.3',
          'Cython>=0.27.3',
          'pandas>=0.20.3',
          'gensim>=3.2.0',
          'scikit_learn>=0.19.1',
      ],
      )
